/************************************************************************

 Module:      JTVRemote.cpp

 Project:     JTV / JTVRemote

 Description: Leadtek Winfast TV2000 FM remote control driver.
 
 Copyright (C) 2002  Ariel Dembling  (adapted the GIRDER driver for JTVRemote)

  Code based on previous work:

  Copyright (c) 2000  Ivan Kolin  (early non-working version of driver for JTVRemote)
  Copyright (C) 2001  Juan Toledo (master copier and paster, Winfast TV2000 plugin for GIRDER)
  Copyright (C) 2001  Ron Bessems (general plugin code, Winfast TV2000 plugin for GIRDER)
  Copyright (C) 2001  John Adcock (BT8x8 driver, Winfast TV2000 plugin for GIRDER)
  Copyright (C) 2001  Avery Lee   (BT8x8 driver initialization, Winfast TV2000 plugin for GIRDER)
  Copyright (C) 2001  Michal Milczewski (polling code, Winfast TV2000 plugin for GIRDER)


 All rights reserved.

 You may not use any piece of the following code for commercial purposes
 without written permission from the author.

 You may freely copy and distribute the source code as long as it contains
 this copyright notice and the disclaimer. You may freely modify it.

 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

************************************************************************/


#include <unknwn.h>
#include <stdio.h>

#include "jtvrem.h"
#include "jtvbt8x8.h"
#include "btregs.h"
#include "jtvremsk.h"

// This driver corresponds to the JTVRemote Specification 1.2.
// Look at JTVRemSpec1x.zip file for additional information.

// Key codes from: Ariel Dembling
// For: "Leadtek Winfast TV2000 FM" with "CoolCommand RM-0010"
// Firm    : Leadtek Research Inc., http://www.leadtek.com.tw
//
// Developed and tested with the following:
// Kit Type: PAL M/N
// Model   : WinFast TV2000 FM
// Tuner   : Philips NTSC
// System  : PAL-NC (Argentina, Paraguay and Uruguay)
//
// Codes obtained by monitoring existing applications with BT878RegTool
// (also available from the LIRC project).
//
// Some other miscellaneous info:
// GPIO_DMA_CTL when TV window open:   0xc007
// GPIO_DMA_CTL when TV window closed: 0xc004
// GPIO_OUT_EN when TV window open or closed:    0xb33c07
//         (that is, bit mapped 10110011 00111100 00000111)
// GPIO_DATA when TV window open:      0xSSMMKK ...
//         ... where: SS = 0x02, 0x22, 0x21 or 0x80 (for STEREO, MONO, SAP or OFF)
//                    MM = 0x2c or 0x0c (for TUNER or COMPOSITE/SVIDEO)
//                    KK = last pressed key code, masked by 0xf8 (see keys[] array definition below)
//



// Definition of key codes:
KeyDef keys[]=
{
// 1-st row keys layout
        // No key
        // No key
        // No key
        { 0x00, "Key_Power", "Power", },
// 2-nd keys row
        { 0x28, "Key_1", "1", },
        { 0x30, "Key_2", "2", },
        { 0x38, "Key_3", "3", },
        { 0x10, "Key_TV/FM", "", },
// ...
        { 0x48, "Key_4", "4", },
        { 0x50, "Key_5", "5", },
        { 0x58, "Key_6", "6", },
        { 0xa8, "Key_AutoScan", "", },
//...
        { 0x68, "Key_7", "7", },
        { 0x70, "Key_8", "8", },
        { 0x78, "Key_9", "9", },
        { 0xb0, "Key_Display", "", },
//...
        { 0x88, "Key_Recall", "Recall", },
        { 0x90, "Key_0", "0", },
        { 0x98, "Key_Enter", "Enter", },
        // No key
//...
        { 0xf8, "Key_CC", "", },
        { 0xc8, "Key_Left", "Left", },
        { 0xc0, "Key_Right", "Right", },
        { 0xf0, "Key_Video", "", },
//...
        { 0xd8, "Key_MTS", "", },
        // No key
        // No key
        { 0xa0, "Key_Mute", "", },
//...
        { 0x60, "Key_ChannelUp", "ChannelUp", },

        { 0x40, "Key_VolumeDown", "VolumeDown", },
        { 0x18, "Key_FullScreen", "", },
        { 0x20, "Key_VolumeUp", "VolumeUp", },

        { 0x80, "Key_ChannelDown", "ChannelDown", },

};




typedef UINT (CALLBACK* LPFNDLLFUNC1)(DWORD,UINT);


DWORD dwParam1;
UINT  uParam2, uReturnVal;

INT first_delay=250;    // default first delay for a key
INT repeat_delay=10;    // default repeat delay for a key


void log(const char* str, ...)
{
  va_list arglist;

  FILE* l=fopen("JTVRemote.log","a+t");
  if (l==NULL) return;
  va_start( arglist, str );
  vfprintf(l, str, arglist );
  va_end( arglist );
  fclose(l);

}




class JTVRemote : public CRemoteDriver
{

public:

		// Constructor
        JTVRemote(KeyDef *k, int count) : CRemoteDriver(k, count)
                {

                log("> Initializing Leadtek Winfast TV2000 FM remote module.\n");
                log("> Copyright (c) 2000 by Ivan Kolin, <ivkolin@mail.ru>, et al.\n");

				// Look at section [JTVRemote] in "jtv.ini" file, look for strings
				// "KeyFirstDelay=" and "KeyRepeatDelay=", for example:
				//      [JTVRemote]
				//      ............
				//      KeyFirstDelay=250
				//      KeyRepeatDelay=10
				// If strings exist, get corresponding values into "first_delay"
				// and "repeat_delay" variables. If not, default values are kept.

				first_delay=GetPrivateProfileInt("JTVRemote","KeyFirstDelay",first_delay,".\\jtv.ini");
				repeat_delay=GetPrivateProfileInt("JTVRemote","KeyRepeatDelay",repeat_delay,".\\jtv.ini");

                log("> KeyFirstDelay=%d, KeyRepeatDelay=%d\n",first_delay,repeat_delay);

				};


		// This method checks for new remote input. I (Ariel) got the idea and part of the
		// code from the GIRDER plugin. I found that key releases on the RC unit are not 
		// informed (only key presses are), and key codes are not repeated when the key is
		// kept pressed. So there seems to be NO way to implement an auto-repeat feature.

		bool IsRemoteDataReady(IBT8x8Device *bt)
		{
			// Note from the GIRDER igirder.cpp driver for the Winfast:
			// "As far as I know, there are two ways to check for remote data readiness with 
			// BT8x8 cards. Checking a flag bit in the GPIO_DATA register or (for the
			// Winfast TV2000 at least) checking for the INT_GPIO bit in the interrupt status
			// register if GPINTC has previously been set." I take the second option here,
			// because I couldn't find the flag bit in GPIO_DATA on this card.


			// From the BT878 documentation, about the BT8x8_INT_STAT register:
			// "This register provides status of pending interrupt conditions. To clear the 
			// interrupts, read this register, then write the same data back. A 1 in the 
			// write data clears the particular register bit. The interrupt/status bits can 
			// be polled at any time."

			DWORD int_stat_value;

			bt->ReadBT(BT848_INT_STAT, &int_stat_value);

			// Is new data ready?
			if (int_stat_value & BT848_INT_GPINT)
			{
				// Yes it is, so we neet to clear the interrupt bit
				bt->WriteBT(BT848_INT_STAT, BT848_INT_GPINT);
				return true;
			}

			return false;
		}

        void PollingThread()
        {
                IBT8x8Device *bt;

                if (SUCCEEDED(device->QueryInterface(IID_IBT8x8Device, (void **)&bt)))
                {
                    INT key_pause=20; // initial repeat delay
                    DWORD test_code=0; // for debugging purposes
		            DWORD value; // will hold key code inputs...
                    DWORD gpio_data=0; // ...extracted from the GPIO_DATA


					// While we don't get busted...
                    while (WaitForSingleObject(hBreakEvent, 0) != WAIT_OBJECT_0)
                    {

                       // Initialize the GPINTC bit, so that each keypress will raise the
	  				   // GPINT bit on the INT_STAT register.
					   // We do this on every cycle, because some applications tend to
					   // reset the wrong bits (BorgTV, for example).
		               bt->ReadBT(BT848_GPIO_DMA_CTL, &value);
			           bt->WriteBT(BT848_GPIO_DMA_CTL, value | BT848_GPIO_DMA_CTL_GPINTC);

					   // We need to sleep before polling, so that we don't use much CPU time
					   Sleep(key_pause);

					   // We test whether there is new data from the remote
					   if (IsRemoteDataReady(bt)) {

							// We read the new key code (the mask 0xf8 is Winfast-specific)
							bt->ReadBT(BT848_GPIO_DATA, &gpio_data);
							value = (gpio_data & 0x000000f8);

							// Test block begin - comment out if building a production version
							//if (test_code != value)
							//{
							//	log(">> ---- Test information ---- \n");
							//	log(">>      Tick_Counter: %d\n",GetTickCount());
							//	log(">>         GPIO_DATA: %x\n",gpio_data);
							//	log(">>  Current_Key_Code: %x\n",value);
							//	test_code = value;
							//}
							// end of Test block

							// Call for related action.
							AddStroke(value);

							// We need to shorten the delay in case the key presses keep coming...
							key_pause = repeat_delay;

                       }
					   else
					   {	
							// There's no new data coming from the remote, so we 
							// slowly enlarge the delay.
							if ( first_delay > (key_pause + repeat_delay)) {
								key_pause = key_pause + repeat_delay;
							}
							else
							{
								key_pause = first_delay;
							};

					   }; // endif new data

					 }; // end while

                log("> Break event received. Bye!\n");

                bt->Release();
                };
        };
};

extern "C"
{
        __declspec(dllexport) IJTVRemoteControl *__stdcall GetInterfacePtr()
        {
                return new JTVRemote(keys, sizeof(keys) / sizeof(keys[0]));
        };
};
